

/* Application file with the stuff called from the callback functions */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>

#include "defs.h"
#include "interface.h"
#include "dlgbox.h"
#include "application.h"
#include "imgbuf.h"
#include "imgmath.h"
#include "magnifier.h"
#include "dicom.h"
#include "roi.h"
#include "measure.h"
#include "functions.h"
#include "statop.h"
#include "tiffstuff.h"
#include "graphdsp.h"
#include "secondwnd.h"
#include "macro.h"
#include "viz.h"



static int statusbar_was_used = 0;
static int segv_trap = 0;

/***************************************************************************/



/********************************************************

	app_init initializes all relevant global variables

*********************************************************/

gboolean watchdog (gpointer data)
{
	/* if (watchdogflag) fprintf (stderr, "%d\n", roicnt);
	if (roicnt>10) watchdogflag=0;
	*/
	return TRUE;
}


void app_init ()
{
int i;

	cimage_pid = getpid();
	strcpy (version_string, "Crystal Image 0.9.8 Beta (Build 062415)");
		/* Note - if you change the version string, make sure it starts
			with "Crystal Image", or adjust the about dialog */

	segv_trap = 0;
	autosize = 1;
	zoomlevel_x=0;
	zoomlevel_y = 0;
	master_abort = 0;
	master_enable = 1;
	autoscale = 1;
	tiffreport=0;
	show_rulers=0;
	tiffsave_dpi = 1200.0;
	dicom_log_string=NULL;
	debuglevel=0;
	skip_license=0;
	cur_slice=0;
	next_uid=1;
	ncontours = 12;
	zslice_callback = NULL;
	file_ok = 0;
	have_magnifier = 0;
	have_thumbnail = 0;
	secondaryimage_exists = 0;
	graphwindow_exists = 0;
	stat_window_exists=0;
	spider_active=0;
	snake_active=0;
	wedgecalib.calibrated=0;
	clickstack_last = 0;
	display_clickstack=0;
	for (i=0; i<CLICKSTACK_MAX; i++) clickstack[i] = (vertex3d){0,0,0};
	spl_cache_invalidate = 1;
	
	intp_pref = INTP_BILIN;		/* By default, we interpolate bilinear */
	browspath[0]=0;
	helppath[0]=0;
	dsopath[0]=0;
	undo_path[0]=0;
	gnuplotpath[0]=0;
	strcpy (imager_sv_path, "/usr/local/bin/cimage_sv");		/* If set up correctly, this should actually work */
																/* TODO: Perhaps, we should check for its existence? */
	imager_cwd = NULL;
	undo_function = U_NONE;
	undo_level=0;
	undo_cnt=0;
	undo_stack = calloc (UNDO_MAX, sizeof (char*));
	for (i=0; i<UNDO_MAX; i++)
		undo_stack[i] = calloc (64,sizeof(char));	/* max 64 bytes in file name */
	
	/* Also reset our global image, secondary image, and math stack */
	
	mainimg.xmax = mainimg.ymax = mainimg.zmax = 0;
	mainimg.imgtype = B_CHAR;
	mainimg.maxval.imax=0;
	mainimg.minval.imin=0;
	mainimg.data=NULL;
	mainimg_name[0]=0;
	mainimg.parent=NULL;
	mainimg.pixbuf=NULL;
	mainimg.drawable=NULL;
	backupimg = mainimg;
	undo_mem = mainimg;
	s_img = mainimg;

	im_stack.x=mainimg;
	im_stack.y=mainimg;
	im_stack.z=mainimg;
	im_stack.t=mainimg;
	
	/* Initialize LUT to neutral */
	
	use_lut=0;
	reset_colormap();
	init_cmap_lut ();
	cmap_window_open = 0;
	viz_bright=0.0; viz_contrast=1.0; viz_gamma=1.0;
	cmap_use_contours = 0; cmap_use_logviz = 0;

	/* Initialize the GSL components */
	
	init_gsl_fn ();

	
	/* Start the ROI marching ants timer - it will be disabled internally
	   by use of a flag */
	 
	roicnt=0;
	dragroi_ini=0;
	watchdogflag=1;
	g_timeout_add (100, (GtkFunction)watchdog, NULL);
	roi_timer_handle = g_timeout_add (300, (GtkFunction)roi_marchingants, NULL);

	/* Zero the macro arrays */
	
	for (i=0; i<MACMAX; i++)
	{
		macromenutbl[i]=NULL;
		macroopts[i]=i;
	}
	nummacros=0;
	macrorun=0;
	for (i=0; i<100; i++) mac_user[i]=0.0;

	/* for the image windows, create a crosshair cursor.
		Find cursor types in gdk/gdkcursors.h.
	    For Fedora systems, the GDK_CROSSHAIR was good, for Kubuntu, use GDK_TARGET */
	
	crosshair_cursor = gdk_cursor_new (GDK_TARGET);
	/* crosshair_cursor = gdk_cursor_new (GDK_CROSSHAIR); */
	/* grab_cursor = gdk_cursor_new (GDK_CIRCLE); */
	grab_cursor = gdk_cursor_new (GDK_HAND2);

}


/* Cleanup, then exit the gtk app */

void app_exit ()
{
int err,i;
char fn[1024];

	/* If we have files in the undo stack, try to delete them */
	
	if ((undo_function==U_DISK) && (undo_level>0))
	{
		for (i=0; i<undo_level; i++)
		{
			if (strlen (undo_path)>0)
			{
				strcpy (fn, undo_path);
				if (fn[strlen(fn)-1]!='/') strcat (fn, "/");
				strcat (fn, undo_stack[i]);
			}
			else
				strcpy (fn, undo_stack[i]);
			err = remove (fn);
			if (err) perror ("Unable to remove temp file");
		}
	}

	gtk_exit (0);
}


/* Trap the SIGKILL and SIGINT interrupts.
	What we are really interested in is to trap
	SIGSEGV - much better to get a nice notice
	than have the program simply "disappear" */


/* SIGSEGV is 11, SIGTERM is 15, and SIGINT is 2. SIGHUP is 1. 
	Just FYI.
*/

void sigkill_handler (int signalnum )
{
	fprintf (stderr, "Kill: Terminated on signal %d.\n", signalnum);
	_exit(0);
}


void sigsegv_handler (int signalnum )
{
	fprintf (stderr, "Oops - Segmentation Violation. Terminating on signal %d.\n", signalnum);
	if (!segv_trap)
	{
		segv_trap=1;
		ShowDynMessage ("Problem: Received SEGV - segmentation fault.\nThis indicates a serious problem.\nPlease restart this program.","CImage: Received SIGSEGV");
	}
	else
	{
		usleep (5000000);
		_exit(0);
	}
}


int int_setup ()
{
static struct sigaction kill_action;
static struct sigaction segv_action;


	/* Trap SIGINT and SIGKILL interrupts */
	
	kill_action.sa_handler = (void (*)(int))sigkill_handler;
	if ( sigaction (SIGINT, &kill_action, NULL ) < 0 ) 
	{
		fprintf (stderr, "Error sigaction() SIGINT\n");
		return -1;
	}
	if ( sigaction (SIGTERM, &kill_action, NULL ) < 0 ) 
	{
		fprintf (stderr, "Error sigaction() TERM\n");
		return -1;
	}

	/* Trap SIGSEGV interrupt */
	
	segv_action.sa_handler = (void (*)(int))sigsegv_handler;
	if ( sigaction (SIGSEGV, &segv_action, NULL ) < 0 ) 
	{
		fprintf (stderr, "Error sigaction() SIGSEGV\n");
		return -1;
	}


	return 0;
}




/*******************************************************

	Add a vertex to the click stack
	
********************************************************/

void clickstack_add (int x, int y, int z)
{
int i;

	if (clickstack_last>=CLICKSTACK_MAX)
	{
		for (i=1; i<=clickstack_last-1; i++)
			clickstack[i-1] = clickstack[i];
		clickstack[CLICKSTACK_MAX-1] = (vertex3d){x,y,z};
	}
	else
	{
		clickstack[clickstack_last++] = (vertex3d){x,y,z};
	}
}

/*******************************************************

	Set the title of the image display (display filename)

*******************************************************/

void set_display_title (char* filename)
{
int i;
char s[255];
char s1[255];

	i = strlen (filename);
	while ((i>0) && (filename[i]!='/')) i--;
	if (filename[i]=='/')
		strcpy (s, &filename[i+1]);
	else
		strcpy (s, filename);
		
	sprintf (s1, "Crystal Image Display (%s)", s);

  	gtk_window_set_title (GTK_WINDOW (mainimg.parent), s1);

}


/****************************************************************

	gtk does not allow us to specifically set adjustment
	values at run-time, so I copied the "gtk_adjustment_init"
	code and modified it accordingly. We NEED to be able to
	modify the Z-slice spinbutton at run-time. Gah.
	
****************************************************************/


void gtk_adjustment_init_val (GtkAdjustment *adjustment, float curval, 
								float lowbnd, float upbnd)
{
	if (curval < lowbnd) curval = lowbnd;
	if (curval > upbnd) curval = upbnd;

	adjustment->value = curval;
	adjustment->lower = lowbnd;
	adjustment->upper = upbnd;
	
	/* Leave the rest unchanged ----
	
  	adjustment->step_increment = 0.0;
  	adjustment->page_increment = 0.0;
  	adjustment->page_size = 0.0;
	
	*/
}



 
/************************************************************

	Save an image file through gdk (always 3-byte RGB format)


*************************************************************/



int save_file (char* filename, image_type* my_im)
{
char *fmtptr;
gint success;
GError *error = NULL;


	/* If there is no image loaded, exit */
	
	if ((!file_ok) || (!my_im)) return -1;

	/* set the image format to be the format of the extension of the filename argument */

	fmtptr = strrchr(filename, '.');
	if (fmtptr)
	{
		fmtptr++;
		success = gdk_pixbuf_save (my_im->pixbuf, filename, fmtptr, &error, NULL);
	}
	else					/* Defaults to PNG */
		success = gdk_pixbuf_save (my_im->pixbuf, filename, "png", &error, NULL);

	if (success)
		return 0;

	if (error) dp (1, "%s\n", error->message);

	return 1;
	
}




/************************************************************


	This is the part which is called from the "File->Open" menu .
	A special helper function exists for DICOM


*************************************************************/


int internal_dicom_read (char* filename, image_type* img)
{
IMAGE* dicom_img;
FILE* dicm_stream;
int dicom_err, numimages;
char dicom_log[16384];


	strcpy (dicom_log, "DICOM LOAD\n");
	dicom_log_string = dicom_log;
	if (debuglevel==0)
		dicom_log_level=ERROR;
	else if (debuglevel==1)
		dicom_log_level=NOTICE;
	else if (debuglevel==2)
		dicom_log_level=INFO;
	else 
		dicom_log_level=DEBUG;
	
	dicm_stream = fopen (filename, "rb"); 
	dicom_init (dicm_stream, dicom_log);
	dicom_err = dicom_read (filename, &dicom_img, &numimages, 1);
	if (dicom_err) 
		dp (0, "Dicom read failed with error %d\n", dicom_err);
	else
	{
		dp (1, "DICOM image successfully read. Size %d x %d x %d, format %d\n",
			dicom_img->w, dicom_img->h, dicom_img->frames, dicom_img->rgb);

		/* if (file_ok) close_file(); -- this is handled elsewhere */
		/* freebuf (img);		-- don't do this here. It will be freed with the backupimg in file_load */
		
		if (dicom_img->rgb)		/* Dicom-RGB */
		{
/*			if (allocate_image (img, B_RGB, dicom_img->w, dicom_img->h, dicom_img->frames)==-1)
				fatal_memerr();
			memcpy (img->data, dicom_img->data.gray, img->zmax*img->xmax*img->ymax*sizeof (char)*3);*/
			img->imgtype = B_RGB;
			img->xmax = dicom_img->w;
			img->ymax = dicom_img->h;
			img->zmax = dicom_img->frames;
			img->data= dicom_img->data.rgb;
		}
		else					/* DICOM-Grayscale 16-bit */
		{
/*			if (allocate_image (img, B_SHORT, dicom_img->w, dicom_img->h, dicom_img->frames)==-1)
				fatal_memerr();
			memcpy (img->data, dicom_img->data.gray, img->zmax*img->xmax*img->ymax*sizeof (short));*/
			img->imgtype = B_SHORT;
			img->xmax = dicom_img->w;
			img->ymax = dicom_img->h;
			img->zmax = dicom_img->frames;
			img->data= dicom_img->data.gray;
		}

		img->uid = next_uid++;

		/* We are done with the dicom buffer, so discard */
		/* dicom_free (dicom_img, 1); */
	}
	if (dicm_stream) fclose (dicm_stream);
	if (tiffreport) ShowLongMessage (dicom_log, "DICOM Log");

	return dicom_err;
}



int internal_file_read (char* filename, image_type* img)
{

int tiff_err;
char success, has_alpha;
int xmax, ymax, stride;
int x, y, idx, jdx;
char filetype;
unsigned char *im_buf;
rgbtriplet *prgb;
GError *gerror;
image_type backupimg;


	success=0;
	memcpy (&backupimg, img, sizeof (backupimg));		/* Back up some values, such as window pointers */
	
	filetype = get_filetype (filename);				/* Try to read & interpret file header */
	switch (filetype)
	{
		case FTYPE_TIFFI:							/* TIFF in Intel or Motorola formats */
		case FTYPE_TIFFM:
				tiff_err = tiffread (img, filename);
				if (!tiff_err)
					success=1;
				break;
		case FTYPE_VSRT:							/* Our own vsrt/vflt image volume formats */
		case FTYPE_VFLT:
				tiff_err = vsrtread (img, filename, (filetype==FTYPE_VFLT));
				if (!tiff_err)
					success=1;
				break;
		case FTYPE_PGM:
		case FTYPE_PPM:
				tiff_err = pnm_read (img, filename);
				if (!tiff_err)
					success=1;
				break;
		case FTYPE_DCM:
			dp (1, "Reading DICOM file %s...\n",filename);
			if (!internal_dicom_read (filename, img)) success=1;
			break;
		default:
				success=0;
	}


	/* Fallback - none of the above applied */
	
	if(!success)	/* Use gdk_pixbuf to open the file. Always defaults to RGB */
	{
		gerror = NULL;
		mainimg.pixbuf = gdk_pixbuf_new_from_file (filename, &gerror);
		if (!mainimg.pixbuf)
		{
			dp (1 , "Error loading file through pixbuf fallback\n");
			success=0;
		}

		if (mainimg.pixbuf)
		{
			/* Get the image's original width and height out of the Image structure, */
			
			/* IMPORTANT: Do not touch w,h afterwards as im->width and
			   im->height are dynamically adjusted to the display size */

			xmax = gdk_pixbuf_get_width (mainimg.pixbuf);
			ymax = gdk_pixbuf_get_height (mainimg.pixbuf);
			stride = gdk_pixbuf_get_rowstride (mainimg.pixbuf);
			has_alpha = gdk_pixbuf_get_has_alpha (mainimg.pixbuf);
			im_buf = gdk_pixbuf_get_pixels (mainimg.pixbuf);			/* Pointer to pixbuf's rgb data */

			dp (1, "Opened file through gdk_pixbuf (%d x %d pixels): %s\n", xmax,ymax,filename);
	
			/* The next step is to transfer the stuff from the imlib buffer to
			     our main buffer to synch them */
				 
			if (allocate_image (img, B_RGB, xmax,ymax,1)==-1) /* a 2D image */
			{
				fatal_memerr ();
			}
			
			img->maxval.imax = 0; img->minval.imin=255; 	/* Sic! */
			prgb = img->data;
			for (y=0; y<ymax; y++)
			{
				idx = stride*y;
				jdx = xmax*y;
				for (x=0; x<xmax; x++)
				{
					prgb[jdx].red = im_buf[idx];
					img->maxval.imax = MAX(img->maxval.imax, im_buf[idx]);
					img->minval.imin = MIN(img->minval.imin, im_buf[idx]);
					idx++;
					prgb[jdx].green = im_buf[idx];
					img->maxval.imax = MAX(img->maxval.imax, im_buf[idx]);
					img->minval.imin = MIN(img->minval.imin, im_buf[idx]);
					idx++;
					prgb[jdx].blue = im_buf[idx];
					img->maxval.imax = MAX(img->maxval.imax, im_buf[idx]);
					img->minval.imin = MIN(img->minval.imin, im_buf[idx]);
					idx++; jdx++;
					if (has_alpha) idx++;			/* Skip over alpha channel */
				}
			}

			img->parent = backupimg.parent;
			img->drawable = backupimg.drawable;
			img->pixbuf = backupimg.pixbuf;
			img->width = backupimg.width; img->height = backupimg.height;

			success=1;			
		}
	}

	if (success) 	reset_colormap();				/* Reset the color map */

	return success;	
	
}





void file_read (char* filename)
{

	/* Save the old image in the backup buffer in case the load op fails */
	
	if (file_ok) backupimg = mainimg; else backupimg.data=NULL;

	statusbar_display ("Loading file...");
	
	if (internal_file_read (filename, &mainimg)>0)
	{

		/* Create an imlib image for display */
		make_minmax (&mainimg);
		cur_slice = 0;
		reset_zoomlevel ();
		make_imlib_buffer_from_main_image (cur_slice);
		prepare_display_window (filename);		/* Causes a configure event followed by an expose event */

		file_ok = 1;
		statusbar_display ("File loaded OK");
		strcpy (mainimg_name, filename);
		
		/* We have successfully opened a new image, now clear the old pointer */
		
		freebuf (&backupimg);

	}		
	else
	{
		/* This leaves the main imge untouched, so clear the backup */
		
		backupimg.data = NULL;
		ShowMessage ("File open failed");
	}

	/* At this point, adjust the Z-slice spin button not to exceed the
	   number of slice we have read. This also syncs it with the 
	   slice 0 being displayed */
	   
	/* wedgecalib.calibrated=0; */

	redraw (0);			/* Just in case, schedule an expose event */
	
}


void prepare_display_window (char* filename)
{
int x,y, mwx, mwy;

	/* Get a good value for the display size.
		Note that we always open the display with full size (TODO: Scaling
		in powers of two), but the parent window may be smaller.
		This is what the scrolledwindow is for. */

	x = mainimg.xmax; y = mainimg.ymax;     /* up to 512*512, keep original image size, zoom factor=1 */
	mainimg.width = x; mainimg.height = y;	/* Display window, always full size */
	mwx = x+33; mwy = y+28;
 	if ((x>640) || (y>512))
 	{
 		mwx = 512;               /* Otherwise, limit to x=512, but retain aspect ratio */
 		mwy = (int) ( (float)x*(float)mainimg.ymax/(float)mainimg.xmax);
 	}

	current_roi.roi_mode=ROI_NONE;
	if (mainimg.parent==NULL)  
		open_imagewindow (&mainimg, mwx, mwy);
	else
		gtk_window_set_default_size (GTK_WINDOW(mainimg.parent), mwx, mwy);
	gtk_widget_show (mainimg.parent);

	set_display_title (filename);
	gdk_window_set_cursor (mainimg.drawable->window, crosshair_cursor);
	dp (2, "Drawing area 1 created. Pointer is %08X, window is %08X\n",mainimg.drawable, mainimg.drawable->window);

	/* A few more adjustments, in case we have a magnifier or thumbnail activated */

	cur_slice=0;
	reset_zoomlevel ();
	if (have_thumbnail)
		display_thumbnail();
	if (have_magnifier)
		destroy_magnifier();		/* Since the mag's center coords may be outside of the image area... */

	gtk_adjustment_init_val ((GtkAdjustment*)slice_spinbutton_adj, 0, 0, mainimg.zmax-1);
}


/* Re-aspect resizes the main display to match the original aspect ratio
   if incdec=0, the size will be reduced.
	THIS IS DEPRECATED AND SHOULD NOT BE USED. */

void re_aspect (char incdec)
{
float fx,fy,f;
int x,y;

	x = mainimg.width;
	y = mainimg.height;
	fx = (float)x/(float)mainimg.xmax;
	fy = (float)y/(float)mainimg.ymax;
	
	if (incdec)	/* increase size of side that is too small */
	{
		if (fx>fy) f=fx; else f=fy;
	}
	else		/* decrease size of side that is too long */
	{
		if (fx<fy) f=fx; else f=fy;
	}
	
	x = (int)(0.5+mainimg.xmax*f);
	y = (int)(0.5+mainimg.ymax*f);
	///gtk_widget_set_size_request (mainimg.drawable, x,y);
	redraw (0);
}


void reset_zoomlevel ()
{
	zoomlevel_x = 0;
	zoomlevel_y = 0;
	gtk_adjustment_set_value (GTK_ADJUSTMENT(zoom_adj), 0);
}


/********************************************************

	close_file closes the display window and discards
	memory used by imlib and main_img

*********************************************************/


void close_file ()		/* Clean up and discard unused memory */
{
	if (file_ok)
	{
		dp (2, "File close\n");

		kill_roi();			/* Must be first to stop timer */

		gtk_widget_hide ( GTK_WIDGET ( mainimg.parent ) );
		redraw (0);		/* Force the system to act in the HIDE above */

		freebuf (&mainimg);	/* discard the image buffer */
		if (mainimg.pixbuf) g_object_unref (G_OBJECT(mainimg.pixbuf));		/* If reference count drops tpo zero, pixbuf will be destroyed */

		mainimg.pixbuf = NULL;

		file_ok = 0;
		reset_colormap();				/* Reset the color map */
		
		if (have_magnifier)
		{
			destroy_magnifier();
			have_magnifier = 0;
		}

		if (have_thumbnail)
		{
			destroy_thumbnail();
			have_thumbnail = 0;
		}

		if (secondaryimage_exists)		/* destroy the secondary image */
			close_secondary_image ();
	
		/* wedgecalib.calibrated=0; */
		
		if ((undo_function==U_MEMORY) && (undo_mem.data))
		{
			freebuf (&undo_mem);
		}
		
	}
}



/********************************************************

	redraw updates the display window from the pixbuf buffer.
	redraw only acts on the imlib image. After any processing step, the
	processing function itself is responsible for updating the pixbuf
	(normally through functions in viz.c)

	Redraw is actually a function to create an expose event.
	NOTE that autosize is gone (for now) and has been replaced
	by a zoom function.
	"must_apply" has been deprecated. There are no agruments to
	redraw().
	
***********************************************************/


void redraw (int must_apply)
{
int w,h;

	if (mainimg.pixbuf)
	{
		
		if (mainimg.drawable->window == NULL)
		{
			dp (0, "Error: Redraw with NULL drawable\n");
			return;
		}

		/* Invalidate the drawing area and schedule an expose event */

		w = mainimg.width;
		h = mainimg.height;
		dp (2, "Redraw: Scheduled expose event\n");
		gtk_widget_queue_draw_area (mainimg.drawable, 0, 0, w, h);

		/* Update magnifier and thumbnail if they exist */
		
		if (have_magnifier)
			display_magnified (&mainimg, mag_lastx, mag_lasty, cur_slice);

		if (have_thumbnail)
			display_thumbnail();

	}
}


		

/* Run through the entire image and extract min and max values */

void make_minmax (image_type* img)
{
int min,max,i;
double fmin,fmax;
int bitmask;
long maxelems;

	min=0, max=255; bitmask=0; fmin=0; fmax=0;
	maxelems = img->xmax * img->ymax * img->zmax;
	if (img->imgtype == B_RGB) maxelems *= 3;
	switch (img->imgtype)
	{
		case B_CHAR:
		case B_RGB:
		{
			unsigned char* p = img->data;
			min = max = *p;			/* initialize min and max */
			for (i=0; i< maxelems; i++)
			{
				min = MIN (*p,min); max = MAX (*p,max);
				bitmask |= *p;
				p++;
			}
			img->minval.imin=min;
			img->maxval.imax=max;
			break;
		}
		case B_SHORT:
		{
			short* p = img->data;
			min = max = *p;			/* initialize min and max */
			for (i=0; i< maxelems; i++)
			{
				min = MIN (*p,min); max = MAX (*p,max);
				bitmask |= *p;
				p++;
			}
			img->minval.imin=min;
			img->maxval.imax=max;
			break;
		}
		case B_FLOAT:
		{
			float* p = img->data;
			fmin = fmax = *p;			/* initialize min and max */
			for (i=0; i< maxelems; i++)
			{
				if (fmin>*p) fmin=*p;
				if (fmax<*p) fmax=*p;
				p++;
			}
			img->minval.fmin=fmin;
			img->maxval.fmax=fmax;
			break;
		}
	}

	if (img->imgtype==B_FLOAT)
		dp (2, "Done reading. Min-Max are %f ... %f\n",
						fmin,fmax);
	else
		dp (2, "Done reading. Min-Max are %d ... %d, bitmask %04x\n",
					min,max, bitmask);
}





/*****************************************************************

	Undo functions
	
*****************************************************************/

/* undo_nexus is the function that gets called from every
   function that changes the mainimage. The mainimage gets
   stored on a stack (either one-image "stack" in memory
   or multiple images on disk). */


void undo_nexus (image_type* img)
{
int err,i;
char fn[1024];

	if (undo_function == U_MEMORY)
	{
		if (undo_mem.data) freebuf (&undo_mem);		/* Overwrite the previous one */
		if (clone_image (img, &undo_mem)==-1)
		{
			dp (0, "Error creating undo backup - out of memory\n");
		}
	}
	else if (undo_function==U_DISK)
	{
		if (undo_level < UNDO_MAX)	/* Stack not filled - just append */
		{
			sprintf (undo_stack[undo_level], "Img_undo_%04x_%03d.tmp",cimage_pid,undo_cnt);
			undo_cnt++; if (undo_cnt>999) undo_cnt=0;
			
			/* save this image as TIFF */
			
			if (strlen (undo_path)>0)
			{
				strcpy (fn, undo_path);
				if (fn[strlen(fn)-1]!='/') strcat (fn, "/");
				strcat (fn, undo_stack[undo_level]);
			}
			else
				strcpy (fn, undo_stack[undo_level]);
			
			err=tiffwrite (fn, img, 0, "cimage UNDO file", 0, img->zmax-1);
	
			undo_level++;
		}
		else		/* Stack full - need to scroll by one image */
		{
		
			/* Delete the oldest image in the stack */
			
			if (strlen (undo_path)>0)
			{
				strcpy (fn, undo_path);
				if (fn[strlen(fn)-1]!='/') strcat (fn, "/");
				strcat (fn, undo_stack[0]);
			}
			else
				strcpy (fn, undo_stack[0]);
			err = remove (fn);
			if (err) perror ("Unable to remove temp file");
			
			/* Scroll down by one */
			
			for (i=1; i<UNDO_MAX; i++)
				strcpy (undo_stack[i-1], undo_stack[i]);
			
			/* save this image as TIFF */
			
			sprintf (undo_stack[UNDO_MAX-1], "Img_undo_%04x_%03d.tmp",cimage_pid,undo_cnt);
			undo_cnt++; if (undo_cnt>999) undo_cnt=0;
			
			if (strlen (undo_path)>0)
			{
				strcpy (fn, undo_path);
				if (fn[strlen(fn)-1]!='/') strcat (fn, "/");
				strcat (fn, undo_stack[UNDO_MAX-1]);
			}
			else
				strcpy (fn, undo_stack[UNDO_MAX-1]);
			
			err=tiffwrite (fn, img, 0, "cimage UNDO file", 0, img->zmax-1);
		
		}
	}
}


/* Error conditions and/or a change in the undo function 
   force us to invalidate the undo stack */

void undo_stack_invalidate ()
{

	if (undo_mem.data) freebuf (&undo_mem);
	undo_mem.data=NULL;
	undo_level=0;
	undo_cnt=0;

}


/* 
	do_undo is the function that actually performs the "undo" menu command.
*/

void do_undo ()
{
char fn[1024];
int err;


	if ((undo_function == U_MEMORY) && (undo_mem.data))
	{
		dp (1, "UNDOing from memory\n");
		freebuf (&mainimg);
		handback (&mainimg, &undo_mem);		/* Better than clone_image, because it leaves the GUI structure intact */
		cur_slice=0;
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (0);
		undo_mem.data = NULL;	/* so we can't do another undo with the same data */
	}
	else if ((undo_function==U_DISK) && (undo_level>0))
	{
		/* determine file name of last image on stack */
	
		if (strlen (undo_path)>0)
		{
			strcpy (fn, undo_path);
			if (fn[strlen(fn)-1]!='/') strcat (fn, "/");
			strcat (fn, undo_stack[undo_level-1]);
		}
		else
			strcpy (fn, undo_stack[undo_level-1]);
		
		/* load the file, then delete it */
		
		if (internal_file_read (fn, &mainimg)==0)
		{
			dp (0, "Undo from disk failed\n");
			undo_stack_invalidate ();	/* if this fails, restart the stack from scratch */
		}

		err = remove (fn);
		if (err) perror ("Unable to remove temp file");
		
		undo_level--;
		if (undo_cnt>0) undo_cnt--;
		cur_slice=0;
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (0);
	}

}


/*****************************************************************

	Display a text message in the control window's status bar
	
*****************************************************************/


void statusbar_display (const char* text)
{

guint context;

	context = gtk_statusbar_get_context_id((GtkStatusbar*)statusbar1, "mycontext");

	if (statusbar_was_used) /* pop the last message */
	{
		gtk_statusbar_pop((GtkStatusbar*)statusbar1, context);
		statusbar_was_used=0;
	}
	
	gtk_statusbar_push((GtkStatusbar*)statusbar1, context, text);
	statusbar_was_used = 1;
}


/*****************************************************************

	Advance or pulse the progress bar; check for abort
	Note that a call to this function disables all GUI functions.
	It is the caller's responsibility to clean up with
	a call to reset_progress ()
	
*****************************************************************/


void indicate_progress (float frac)
{
int master_en_save;
	
	if (frac > 1.0) frac=1.0;

	if (frac<0) 
	{
		gtk_progress_bar_set_pulse_step (GTK_PROGRESS_BAR(progressbar), 0.05);
		gtk_progress_bar_pulse (GTK_PROGRESS_BAR(progressbar));			/* frac<0 indtcates that we don't know what is 100% */
	}
	else
		gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR(progressbar), frac);	/* Set the progressbar */

	/* Now give gtk a chance to catch up; monitor for the stop button */
	/* We must disallow any other function to avoid confusion */

	master_en_save = master_enable;
	master_enable = 0;

	while (gtk_events_pending())
	{
		dp (3, "Progressbar: Handling pending event\n");
		gtk_main_iteration();
	}
	master_enable = master_en_save;
}


void reset_progress ()
{

	master_enable = 1;
	gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR(progressbar), 0.0);	/* Set the progressbar to 0.0 */

}

/**************************************************************+

	Register and unregister a function for the 
	z-slice spinbutton
	Modify the z slice we are looking at
	
**************************************************************/

void zslice_spin_register (gpointer func)
{
	zslice_callback = func;
}


void zslice_spin_unregister ()
{
	zslice_callback = NULL;
}


void new_zslice (int z)
{
	if (z >= mainimg.zmax) z=mainimg.zmax-1;
	else if (z<0) z=0;

	make_imlib_buffer_from_main_image (z);
	redraw(0);
	if (have_magnifier)
		display_magnified (&mainimg, mag_lastx, mag_lasty, z);
	if (have_thumbnail)
		display_thumbnail ();
	if (stat_window_exists)
		statop_redraw_all (0);
	if (zslice_callback)
			zslice_callback ();	/* any registered z-slice callback func */
}




/**************************************************************+

	Init_scrollbars is used to set up the scrollbars
	of the primary display whenever non-autoscaled mode
	is selected.

	DEPRECATED as of version 0.9.0 where scrolling is
	handled very differently.
	
**************************************************************/

void init_scrollbars ()
{

}


void about_window ()
{
char s[255];
char c[255];

	strcpy (c,"About Crystal Image...");
	sprintf (s, "Crystal Image %s\n(c) 2000-2010 Mark A. Haidekker", VERSION);

	ShowMessageC (s,c);
	
}


GtkWidget *display_license ()
{
char s[16384];
char c[255];

	strcpy (c,"Crystal Image License");
	strcpy (s,"This software is copyrighted material.\n"\
"Copyright (C) 2000-2015 by Mark A. Haidekker, all rights reserved.\n"\
"\n"\
"This program is free software: you can redistribute it and/or modify\n"\
"it under the terms of the GNU General Public License as published by\n"\
"the Free Software Foundation, either version 3 of the License, or\n"\
"(at your option) any later version.\n\n"\
"This program is distributed in the hope that it will be useful,\n"\
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"\
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"\
"GNU General Public License for more details.\n\n"\
"You should have received a copy of the GNU General Public License\n"
"along with this program.  If not, see <http://www.gnu.org/licenses/>.\n");

	return ShowDynMessage (s,c);

	
}




gboolean license_timer (gpointer data)
{

	dp (2, "License splash timer elapsed. Destroying window at %08x\n", (int)data);
	if ( GTK_IS_WIDGET (data) )						/* Make sure the user hasn't closed the window */
		gtk_widget_destroy ( GTK_WIDGET ( data ) );	/* Data is a pointer to the license widget */

	return FALSE;		/* Timer does not repeat */
}


void display_license_splash ()
{
GtkWidget *splash;
gint timer_handle;

	splash = display_license();		/* Display the license and get a pointer to the window */
	/* Wait 5 seconds then destroy the window */
	timer_handle = g_timeout_add (5000, (GtkFunction)license_timer, splash);
	dp (2, "License window pointer %08x linked to timer %d\n",(int)splash, timer_handle);
}


int start_help()
{
int pid, rtncode;
char* argv[3];

	if ((strlen(browspath)==0) || (strlen(helppath)==0))
	{
		ShowMessage ("Browser or help path not specified");
		return -1;
	}

	rtncode=0;
	argv[0]= browspath;
	argv[1] =helppath;
	argv[2]=NULL;
	
	dp (1, "Help: Forking...");
	
	pid = fork();
	
	/* This is the first statement executed by the original and the forked process */
	
	dp (1, "   fork PID %d\n", pid);
	if (pid==-1)
	{
		ShowMessage ("Error: Could not start browser");
		return -1;
	}
	
	/* Note that at this point the forked process and the parent
	   process have different values for pid. The parent has the
	   regular pid of the child, thus can perform "waitpid" further down.
	   The child thinks it`s pid==0 and thus enters the if-statement
	   that follows, and terminates on the execl statement */
	
	if (pid==0)
	{

		dp (1, "Cimage (forked): starting backend, pid=%d\n", pid);
		rtncode = execv (browspath, argv);
		
		/* The following statements will not be executed 
		   if execl was successful */
		
		dp (0, "Cimage (forked): This message should not appear. Failure %d\n", rtncode);
		exit (127);
	}

	return pid;		/* pid of the forked process == pid of slave process */

}




/********************************************************************

	clickhandler is called in response to a non-drag click event
	and activates pixelpeek
	
*********************************************************************/

void clickhandler (int mousex, int mousey)
{
int x,y;
int ix, iy;


	ix = mainimg.drawable->allocation.width;
	iy = mainimg.drawable->allocation.height;
	x = mousex; y = mousey;

	roi_imgxy (&x,&y);
	
	if (x>=mainimg.xmax)
	{
		x = mainimg.xmax-1;
		dp (1, "Warning: Out of bound X coordinate adjusted\n");
	}
	if (y>=mainimg.ymax)
	{
		y = mainimg.ymax-1;
		dp (1, "Warning: Out of bound Y coordinate adjusted\n");
	}
	
	clickstack_add (x,y,cur_slice);
	pixelpeek (x,y);
	if (have_magnifier) display_magnified (&mainimg, mousex, mousey,cur_slice);
}



/********************************************************************

	Pixelpeek displays the value under the cursor in the statusbar
	- only in the active slice
	
*********************************************************************/



void pixelpeek (int x, int y)
{
char s[255];
long zz;


	if (!mainimg.data) return;	/* No image, no peeking */
	if ((x<0) || (y<0) || (x>=mainimg.xmax) || (y>mainimg.ymax)) return;	/* Baad - out of range */

	zz = x + mainimg.xmax*(y + mainimg.ymax*cur_slice);
	switch (mainimg.imgtype)
	{
		case B_CHAR: 
		{
			unsigned char* p = mainimg.data;
			if (wedgecalib.calibrated)
				sprintf (s, " (%d,%d,%d) Data=%d Calibrated=%f", x,y,cur_slice,p[zz], calibrate_val(p[zz]));
			else			
				sprintf (s, " (%d,%d,%d) Data=%d", x,y,cur_slice,p[zz]);
			break;
		}
		case B_SHORT: 
		{
			short* p = mainimg.data;
			if (wedgecalib.calibrated)
				sprintf (s, " (%d,%d,%d) Data=%d Calibrated=%f", x,y,cur_slice,p[zz], calibrate_val(p[zz]));
			else			
				sprintf (s, " (%d,%d,%d) Data=%d", x,y,cur_slice,p[zz]);
			break;
		}
		case B_RGB: 
		{
			unsigned char* p = mainimg.data;
			zz *= 3;	/* because we have three chars per pixel */
			sprintf (s, " (%d,%d,%d) R=%d G=%d B=%d",x,y,cur_slice,p[zz],p[zz+1],p[zz+2]);
			break;
		}
		case B_FLOAT: 
		{
			float* p = mainimg.data;
			if (wedgecalib.calibrated)
				sprintf (s, " (%d,%d,%d) Data=%10.4f Calibrated=%f", x,y,cur_slice,p[zz], calibrate_val(p[zz]));
			else			
				sprintf (s, " (%d,%d,%d) Data=%10.4f", x,y,cur_slice,p[zz]);
			break;
		}
	}
	statusbar_display (s);
	
}


/***********************************************+

	View_Profile runs along a straight line
	between the two most recent entries in
	the clickstack and displays the gray values
	
*************************************************/

void view_profile_xyz (image_type* img, int x1, int y1, int z1, int x2, int y2, int z2)
{
double x,y,z;
double dx,dy,dz;
int numpts,i;
float* dist;
float* grayval;
int e;
rgbtriplet rgb;

	e = 1; if (img->imgtype==B_RGB) e=3;

	dx = x2-x1; dy = y2-y1; dz = z2-z1;
	
	if ( (fabs(dx)>=fabs(dy)) && (fabs(dx)>=fabs(dz)) )	/* x-dominant */
		numpts = abs(x2 - x1);
	else if ( (fabs(dy)>=fabs(dx)) && (fabs(dy)>=fabs(dz)) )	/* y-dominant */
		numpts = abs(y2 - y1);
	else 												/* z-dominant */
		numpts = abs(z2 - z1);

	dy /= numpts;
	dz /= numpts;
	dx /= numpts; 	/* This retains the sign */

	dist    = (float*) calloc (numpts, sizeof (float));
	grayval = (float*) calloc (numpts*e, sizeof (float));
	if ((!dist) || (!grayval)) fatal_memerr();
	
	display_clickstack=2;

	x=x1; y=y1; z=z1;

	if (img->imgtype==B_RGB)
	{
		for (i=0; i<numpts; i++)
		{
			/* TODO: Doesn't work across slices with slanted lines */
			
			rgb = rgbireadbuf (img, x,y,(int)(z+0.5),intp_pref);
			grayval[i]=rgb.red;
			grayval[i + numpts] = rgb.green;
			grayval[i + 2*numpts] = rgb.blue;
			dist[i]=i;
			x+=dx; y+=dy; z+=dz;
		}
	
		multi_graphme (dist, grayval, numpts,e, "Relative Position", "Intensity (R, G, B)", "Intensity Profile", 0);
	}
	else
	{
		for (i=0; i<numpts; i++)
		{
			/* TODO: Doesn't work across slices */
			
			grayval[i] = ireadbuf (img, x,y,(int)(z+0.5),intp_pref);
			dist[i]=i;
			x+=dx; y+=dy; z+=dz;
		}
		graphme (dist, grayval, numpts, "Relative Position", "Intensity", "Intensity Profile", 0);
	}
	
	free (dist);
	free (grayval);

}


void view_profile ()
{
vertex3d start, end;

	/* Get the two most recent points that have been clicked */

	if (clickstack_last < 2) return;
	start = clickstack[clickstack_last-2];
	end   = clickstack[clickstack_last-1];
	if ((start.x==end.x) && (start.y==end.y) && (start.z==end.z)) return;

	view_profile_xyz (&mainimg, start.x,start.y,start.z, end.x,end.y,end.z);

}


/* Linesum_x didplays an intensity profile from left to right, averaging columns */

void view_linesum_x (image_type* img, int z)
{
float* xpos;
float* grayval;
int e;
rgbtriplet rgb; rgbtriplet* prgb;
float f,red,green,blue;
int x,y,numpts,numavg;
int rx1,ry1,rx2,ry2;


	roi_bbox (img, &rx1,&ry1,&rx2,&ry2);
	prgb=img->data;

	numpts = rx2-rx1+1;
	numavg = ry2-ry1+1;
	e = 1; if (img->imgtype==B_RGB) e=3;
	xpos    = (float*) calloc (numpts, sizeof (float));
	grayval = (float*) calloc (numpts*e, sizeof (float));
	if ((!xpos) || (!grayval)) fatal_memerr();


	for (x=rx1; x<=rx2; x++)
	{
		red=0; green=0; blue=0; f=0.0;
		if (img->imgtype == B_RGB)
		{
			for (y=ry1; y<=ry2; y++)
			{
				rgb = prgb[x + img->xmax*(y+img->ymax*z)];
				red   += rgb.red;
				green += rgb.green;
				blue  += rgb.blue;
			}
			xpos[x-rx1]=x;
			grayval[x-rx1]=red/numavg;
			grayval[x-rx1 + numpts] = green/numavg;
			grayval[x-rx1 + 2*numpts] = blue/numavg;
		}
		else
		{
			for (y=ry1; y<=ry2; y++)
			{
				f += readbuf_flt (img,x,y,z);
			}
			xpos[x-rx1]=x;
			grayval[x-rx1]=f/numavg;
		}
	}

	if (e==1)
		graphme (xpos, grayval, numpts, "Relative Position", "Intensity", "Averaged Intensity Profile", 0);
	else
		multi_graphme (xpos, grayval, numpts,e, "Relative Position", "Intensity (R,G,B)", "Averaged Intensity Profile", 0);
	
	free (xpos);
	free (grayval);

}


/* Linesum_y didplays an intensity profile from top to bottom, averaging rows */

void view_linesum_y (image_type* img, int z)
{
float* xpos;
float* grayval;
int e;
rgbtriplet rgb; rgbtriplet* prgb;
float f,red,green,blue;
int x,y,numpts,numavg;
int rx1,ry1,rx2,ry2;


	roi_bbox (img, &rx1,&ry1,&rx2,&ry2);
	prgb=img->data;

	numpts = ry2-ry1+1;
	numavg = rx2-rx1+1;
	e = 1; if (img->imgtype==B_RGB) e=3;
	xpos    = (float*) calloc (numpts, sizeof (float));
	grayval = (float*) calloc (numpts*e, sizeof (float));
	if ((!xpos) || (!grayval)) fatal_memerr();


	for (y=ry1; y<=ry2; y++)
	{
		red=0; green=0; blue=0; f=0.0;
		if (img->imgtype == B_RGB)
		{
			for (x=rx1; x<=rx2; x++)
			{
				rgb = prgb[x + img->xmax*(y+img->ymax*z)];
				red   += rgb.red;
				green += rgb.green;
				blue  += rgb.blue;
			}
			xpos[y-ry1]=y;
			grayval[y-ry1]=red/numavg;
			grayval[y-ry1 + numpts] = green/numavg;
			grayval[y-ry1 + 2*numpts] = blue/numavg;
		}
		else
		{
			for (x=rx1; x<=rx2; x++)
			{
				f += readbuf_flt (img,x,y,z);
			}
			xpos[y-ry1]=y;
			grayval[y-ry1]=f/numavg;
		}
	}

	if (e==1)
		graphme (xpos, grayval, numpts, "Relative Position", "Intensity", "Averaged Intensity Profile", 0);
	else
		multi_graphme (xpos, grayval, numpts,e, "Relative Position", "Intensity (R,G,B)", "Averaged Intensity Profile", 0);
	
	free (xpos);
	free (grayval);

}



/********************************************************

	Info: Display info about image
	
********************************************************/


void imginfo (image_type* img)
{
char s[128]; 
char result[4096];
long imgsize;
int bx1,by1,bx2,by2,i;


	strcpy (result, "Information about image:\n------------------------\n");
	if (img->zmax==1)
		sprintf (s, "Dimensions: %li by %li (2D slice)\n",img->xmax,img->ymax);
	else
		sprintf (s, "Dimensions: %li by %li by %li\n",img->xmax,img->ymax, img->zmax);
	strcat (result, s);
	
	sprintf (s, "Image type (data type): ");
	strcat (result, s);
	switch (img->imgtype)
	{
		case B_CHAR:
				strcpy (s, "Unsigned byte, 8 bpp\n");
				break;	
		case B_SHORT:
				strcpy (s, "Signed short, 16 bpp\n");
				break;	
		case B_FLOAT:
				strcpy (s, "IEEE floating-point, 32 bpp\n");
				break;	
		case B_RGB:
				strcpy (s, "RGB, 24 bpp\n");
				break;	
	}
	strcat (result, s);

	if (img->imgtype==B_FLOAT)
		sprintf (s, "Value range: %f to %f\n",img->minval.fmin, img->maxval.fmax);
	else
		sprintf (s, "Value range: %d to %d\n",img->minval.imin, img->maxval.imax);
	strcat (result, s);

	imgsize = img->xmax*img->ymax*img->zmax * DATALEN(img->imgtype);
	if (imgsize <= 4194304)
		sprintf (s, "Total image memory %li KB (%li bytes)\n",imgsize/1024, imgsize);
	else
		sprintf (s, "Total image memory %li MB (%li bytes)\n",imgsize/1048576, imgsize);
	strcat (result, s);

	if (current_roi.roi_mode)	/* Display ROI info as well */
	{
		strcpy (s, "Information about ROI:\n------------------------\n");
		strcat (result, s);

		roi_bbox (&mainimg, &bx1,&by1, &bx2, &by2);
		switch (current_roi.roi_mode)
		{
			case ROI_RECT:				/* Rectangular ROI */
			{
				sprintf (s, "Rectangular ROI from %dx%d to %dx%d\n",
					current_roi.xi[0], current_roi.yi[0], current_roi.xi[1], current_roi.yi[1]);
				strcat (result,s);
				break;
			}
			case ROI_POLY:
			{
				sprintf (s, "Vertex table for polygonal ROI\n");
				strcat (result,s);
				sprintf (s, "with bounding box from %dx%d to %dx%d\n",bx1,by1,bx2,by2);
				strcat (result,s);
				
				for (i=0; i<current_roi.numvertices; i++) 
				{
					sprintf (s, "%d:   %d   %d\n", i,current_roi.xi[i], current_roi.yi[i]);
					if (strlen (result) < 3800) strcat (result,s);
				}

			}
			case ROI_CIRCLE:				/* circular ROI */
			{
				sprintf (s, "Circular ROI at %d,%d, radius %d\n",
					current_roi.xc, current_roi.yc, current_roi.radius);
				strcat (result,s);
				break;
			}
		}
	}
	
	ShowMessageC (result, "Image Info");


}







